/*
 This file is part of the Simple Java Message Exchange (SIJAME) software.

(c) 2011 INSERM
SIJAME is developed at Montpellier RIO Imaging (IFR 122), Montpellier, France (www.mri.cnrs.fr)
Developer: Volker Baecker (volker.baecker@mri.cnrs.fr) 

This software is a very simple message oriented middleware (mom). It allows 
a communication between programms via sockets. A software (client) can send a 
message to another software on another machine (server). The message is 
put into a message queue. Whenever the currently first message in the queue changes
a notification is sent to observers of the queue.  

This software is governed by the CeCILL-B license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-B
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-B license and that you accept its terms. 
 */
package fr.cnrs.mri.sijame.messages;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.logging.Level;

import fr.cnrs.mri.util.logging.LoggingUtil;

/**
 * An observable message queue. A notification is sent to observers 
 * whenever the first message in the queue changes, i.e. when a message
 * is pushed into an empty queue or when the first message is poped from
 * a not empty queue.
 */
public class MessageQueue extends Observable {
	public enum Aspect {CONTENT};
	
	/**
	 * The list of messages that constitutes the queue.
	 */
	private List<Message> elements = new ArrayList<Message>();

	/**
	 * Test whether the queue is empty.
	 * 
	 * @return	True if the queue is empty and false otherwise.
	 */
	public boolean isEmpty() {
		return elements.isEmpty();
	}

	/**
	 * Add a message at the end of the queue. If the queue has been empty before,
	 * the message will be sent to all observers of the queue.
	 * 
	 * @param message	The message that will be added to the queue.
	 */
	public void push(Message message) {
		elements.add(message);
		this.changed(Aspect.CONTENT);
		if (this.size()==1) this.notifyCurrentMessageChanged(message);
	}

	/**
	 * Answer the size of the queue, i.e. the number of messages in the queue.
	 * 
	 * @return	The size of the queue.
	 */
	public int size() {
		return elements.size();
	}

	/**
	 * Answer a the first element of the queue if any. The element will remain in the queue.
	 * 
	 * @return	The first element in the queue.
	 */
	public Message front() {
		if (elements.isEmpty()) return null;
		return elements.get(0);
	}

	/**
	 * Remove the first element from the queue. If the queue is not empty afterwards,
	 * the new first message will be sent to all observers.
	 */
	public void pop() {
		if (elements.isEmpty()) return;
		elements.remove(0);
		this.changed(Aspect.CONTENT);
		if (elements.isEmpty()) return;
		this.notifyCurrentMessageChanged(elements.get(0));
	}

	/**
	 * All messages are removed from the queue. No notifications are sent. 
	 * Observers are not changed.
	 */
	public void reset() {
		this.elements.removeAll(this.elements);
		this.changed(Aspect.CONTENT);
	}

	/**
	 * Send the message to the observers of the queue.
	 * 
	 * @param message	The message that will be sent to observers.
	 */
	private void notifyCurrentMessageChanged(Message message) {
		this.setChanged();
		this.notifyObservers(message);
		this.clearChanged();
		if (message.hasBeenConsumed()) this.pop();
	}
	
	private void changed(Aspect anAspect) {
		this.setChanged();
		this.notifyObservers(anAspect);
		this.clearChanged();
	}

	public void saveAs(String filename) {
		try{
			OutputStream file = new FileOutputStream( filename );
			OutputStream buffer = new BufferedOutputStream( file );
			ObjectOutput output = new ObjectOutputStream( buffer );
			try{
				output.writeObject(elements);
			}
			finally{
				output.close();
			}
		}  
		catch(IOException ex){
			LoggingUtil.getLoggerFor(this).log(Level.SEVERE, "Cannot perform output.", ex);
		}
	}
	
	@SuppressWarnings("unchecked")
	public void load(String filename) {
		if (!new File(filename).exists()) return;
		try{
			InputStream file = new FileInputStream( filename );
			InputStream buffer = new BufferedInputStream( file );
			ObjectInput input = new ObjectInputStream ( buffer );
			try{
				elements = (ArrayList<Message>)input.readObject();
			}
			finally{
				input.close();
			}
		}
		catch(ClassNotFoundException ex){
			LoggingUtil.getLoggerFor(this).log(Level.SEVERE, "Cannot perform input. Class not found.", ex);
		}
		catch(IOException ex){
			LoggingUtil.getLoggerFor(this).log(Level.SEVERE, "Cannot perform input.", ex);
		}

	}

}
